@@ -22,8 +22,6 @@
 #include <string.h>
 #include <time.h>
 
-#include <asm/byteorder.h>
-
 #include "contain.h"
 #include "msg.h"
 #include "print.h"
@@ -80,16 +78,6 @@ static void announce_post_recv(struct announce_msg *m)
 	m->stepsRemoved = ntohs(m->stepsRemoved);
 }
 
-int64_t host2net64(int64_t val)
-{
-	return __cpu_to_be64(val);
-}
-
-int64_t net2host64(int64_t val)
-{
-	return __be64_to_cpu(val);
-}
-
 static int hdr_post_recv(struct ptp_header *m)
 {
 	if ((m->ver & VERSION_MASK) != VERSION)
@@ -174,6 +162,16 @@ static struct tlv_extra *msg_tlv_prepare(struct ptp_message *msg, int length)
 	return extra;
 }
 
+static void msg_tlv_recycle(struct ptp_message *msg)
+{
+	struct tlv_extra *extra;
+
+	while ((extra = TAILQ_FIRST(&msg->tlv_list)) != NULL) {
+		TAILQ_REMOVE(&msg->tlv_list, extra, list);
+		tlv_extra_recycle(extra);
+	}
+}
+
 static void port_id_post_recv(struct PortIdentity *pid)
 {
 	pid->portNumber = ntohs(pid->portNumber);
@@ -184,10 +182,11 @@ static void port_id_pre_send(struct PortIdentity *pid)
 	pid->portNumber = htons(pid->portNumber);
 }
 
-static int suffix_post_recv(struct ptp_message *msg, uint8_t *ptr, int len)
+static int suffix_post_recv(struct ptp_message *msg, int len)
 {
+	uint8_t *ptr = msg_suffix(msg);
 	struct tlv_extra *extra;
-	int err;
+	int err, suffix_len = 0;
 
 	if (!ptr)
 		return 0;
@@ -205,12 +204,14 @@ static int suffix_post_recv(struct ptp_message *msg, uint8_t *ptr, int len)
 			tlv_extra_recycle(extra);
 			return -EBADMSG;
 		}
+		suffix_len += sizeof(struct TLV);
 		len -= sizeof(struct TLV);
 		ptr += sizeof(struct TLV);
 		if (extra->tlv->length > len) {
 			tlv_extra_recycle(extra);
 			return -EBADMSG;
 		}
+		suffix_len += extra->tlv->length;
 		len -= extra->tlv->length;
 		ptr += extra->tlv->length;
 		err = tlv_post_recv(extra);
@@ -220,7 +221,7 @@ static int suffix_post_recv(struct ptp_message *msg, uint8_t *ptr, int len)
 		}
 		msg_tlv_attach(msg, extra);
 	}
-	return 0;
+	return suffix_len;
 }
 
 static void suffix_pre_send(struct ptp_message *msg)
@@ -234,6 +235,7 @@ static void suffix_pre_send(struct ptp_message *msg)
 		tlv->type = htons(tlv->type);
 		tlv->length = htons(tlv->length);
 	}
+	msg_tlv_recycle(msg);
 }
 
 static void timestamp_post_recv(struct ptp_message *m, struct Timestamp *ts)
@@ -294,6 +296,42 @@ void msg_cleanup(void)
 	}
 }
 
+struct ptp_message *msg_duplicate(struct ptp_message *msg, int cnt)
+{
+	struct ptp_message *dup;
+	int err;
+
+	dup = msg_allocate();
+	if (!dup) {
+		return NULL;
+	}
+	memcpy(dup, msg, sizeof(*dup));
+	dup->refcnt = 1;
+	TAILQ_INIT(&dup->tlv_list);
+
+	err = msg_post_recv(dup, cnt);
+	if (err) {
+		switch (err) {
+		case -EBADMSG:
+			pr_err("msg_duplicate: bad message");
+			break;
+		case -EPROTO:
+			pr_debug("msg_duplicate: ignoring message");
+			break;
+		}
+		msg_put(dup);
+		return NULL;
+	}
+	if (msg_sots_missing(msg)) {
+		pr_err("msg_duplicate: received %s without timestamp",
+		       msg_type_string(msg_type(msg)));
+		msg_put(dup);
+		return NULL;
+	}
+
+	return dup;
+}
+
 void msg_get(struct ptp_message *m)
 {
 	m->refcnt++;
@@ -301,8 +339,7 @@ void msg_get(struct ptp_message *m)
 
 int msg_post_recv(struct ptp_message *m, int cnt)
 {
-	int pdulen, type, err;
-	uint8_t *suffix = NULL;
+	int err, pdulen, suffix_len, type;
 
 	if (cnt < sizeof(struct ptp_header))
 		return -EBADMSG;
@@ -356,7 +393,6 @@ int msg_post_recv(struct ptp_message *m, int cnt)
 		timestamp_post_recv(m, &m->sync.originTimestamp);
 		break;
 	case DELAY_REQ:
-		suffix = m->delay_req.suffix;
 		break;
 	case PDELAY_REQ:
 		break;
@@ -366,36 +402,35 @@ int msg_post_recv(struct ptp_message *m, int cnt)
 		break;
 	case FOLLOW_UP:
 		timestamp_post_recv(m, &m->follow_up.preciseOriginTimestamp);
-		suffix = m->follow_up.suffix;
 		break;
 	case DELAY_RESP:
 		timestamp_post_recv(m, &m->delay_resp.receiveTimestamp);
 		port_id_post_recv(&m->delay_resp.requestingPortIdentity);
-		suffix = m->delay_resp.suffix;
 		break;
 	case PDELAY_RESP_FOLLOW_UP:
 		timestamp_post_recv(m, &m->pdelay_resp_fup.responseOriginTimestamp);
 		port_id_post_recv(&m->pdelay_resp_fup.requestingPortIdentity);
-		suffix = m->pdelay_resp_fup.suffix;
 		break;
 	case ANNOUNCE:
 		clock_gettime(CLOCK_MONOTONIC, &m->ts.host);
 		timestamp_post_recv(m, &m->announce.originTimestamp);
 		announce_post_recv(&m->announce);
-		suffix = m->announce.suffix;
 		break;
 	case SIGNALING:
-		suffix = m->signaling.suffix;
+		port_id_post_recv(&m->signaling.targetPortIdentity);
 		break;
 	case MANAGEMENT:
 		port_id_post_recv(&m->management.targetPortIdentity);
-		suffix = m->management.suffix;
 		break;
 	}
 
-	err = suffix_post_recv(m, suffix, cnt - pdulen);
-	if (err)
-		return err;
+	suffix_len = suffix_post_recv(m, cnt - pdulen);
+	if (suffix_len < 0) {
+		return suffix_len;
+	}
+	if (pdulen + suffix_len != m->header.messageLength) {
+		return -EBADMSG;
+	}
 
 	return 0;
 }
@@ -437,6 +472,7 @@ int msg_pre_send(struct ptp_message *m)
 		announce_pre_send(&m->announce);
 		break;
 	case SIGNALING:
+		port_id_pre_send(&m->signaling.targetPortIdentity);
 		break;
 	case MANAGEMENT:
 		port_id_pre_send(&m->management.targetPortIdentity);
@@ -463,7 +499,19 @@ struct tlv_extra *msg_tlv_append(struct ptp_message *msg, int length)
 void msg_tlv_attach(struct ptp_message *msg, struct tlv_extra *extra)
 {
 	TAILQ_INSERT_TAIL(&msg->tlv_list, extra, list);
-	msg->tlv_count++;
+}
+
+int msg_tlv_count(struct ptp_message *msg)
+{
+	int count = 0;
+	struct tlv_extra *extra;
+
+	for (extra = TAILQ_FIRST(&msg->tlv_list);
+			extra != NULL;
+			extra = TAILQ_NEXT(extra, list))
+		count++;
+
+	return count;
 }
 
 const char *msg_type_string(int type)
@@ -529,18 +577,13 @@ void msg_print(struct ptp_message *m, FILE *fp)
 
 void msg_put(struct ptp_message *m)
 {
-	struct tlv_extra *extra;
-
 	m->refcnt--;
 	if (m->refcnt) {
 		return;
 	}
 	pool_stats.count++;
 	pool_debug("recycle", m);
-	while ((extra = TAILQ_FIRST(&m->tlv_list)) != NULL) {
-		TAILQ_REMOVE(&m->tlv_list, extra, list);
-		tlv_extra_recycle(extra);
-	}
+	msg_tlv_recycle(m);
 	TAILQ_INSERT_HEAD(&msg_pool, m, list);
 }
 
