@@ -18,6 +18,7 @@
  */
 #include <arpa/inet.h>
 #include <errno.h>
+#include <stdbool.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -52,6 +53,20 @@ static void scaled_ns_h2n(ScaledNs *sns)
 	sns->fractional_nanoseconds = htons(sns->fractional_nanoseconds);
 }
 
+static void timestamp_host2net(struct Timestamp *t)
+{
+	HTONL(t->seconds_lsb);
+	HTONS(t->seconds_msb);
+	HTONL(t->nanoseconds);
+}
+
+static void timestamp_net2host(struct Timestamp *t)
+{
+	NTOHL(t->seconds_lsb);
+	NTOHS(t->seconds_msb);
+	NTOHL(t->nanoseconds);
+}
+
 static uint16_t flip16(uint16_t *p)
 {
 	uint16_t v;
@@ -79,6 +94,22 @@ static int64_t net2host64_unaligned(int64_t *p)
 	return v;
 }
 
+static size_t tlv_array_count(struct TLV *tlv, size_t base_size, size_t item_size)
+{
+	return (tlv->length - base_size) / item_size;
+}
+
+static bool tlv_array_invalid(struct TLV *tlv, size_t base_size, size_t item_size)
+{
+	size_t expected_length, n_items;
+
+	n_items = tlv_array_count(tlv, base_size, item_size);
+
+	expected_length = base_size + n_items * item_size;
+
+	return (tlv->length == expected_length) ? false : true;
+}
+
 static int mgt_post_recv(struct management_tlv *m, uint16_t data_len,
 			 struct tlv_extra *extra)
 {
@@ -92,6 +123,7 @@ static int mgt_post_recv(struct management_tlv *m, uint16_t data_len,
 	struct grandmaster_settings_np *gsn;
 	struct subscribe_events_np *sen;
 	struct port_properties_np *ppn;
+	struct port_stats_np *psn;
 	struct mgmt_clock_description *cd;
 	int extra_len = 0, len;
 	uint8_t *buf;
@@ -286,6 +318,14 @@ static int mgt_post_recv(struct management_tlv *m, uint16_t data_len,
 		extra_len = sizeof(struct port_properties_np);
 		extra_len += ppn->interface.length;
 		break;
+	case TLV_PORT_STATS_NP:
+		if (data_len < sizeof(struct port_stats_np))
+			goto bad_length;
+		psn = (struct port_stats_np *)m->data;
+		psn->portIdentity.portNumber =
+			ntohs(psn->portIdentity.portNumber);
+		extra_len = sizeof(struct port_stats_np);
+		break;
 	case TLV_SAVE_IN_NON_VOLATILE_STORAGE:
 	case TLV_RESET_NON_VOLATILE_STORAGE:
 	case TLV_INITIALIZE:
@@ -319,6 +359,7 @@ static void mgt_pre_send(struct management_tlv *m, struct tlv_extra *extra)
 	struct grandmaster_settings_np *gsn;
 	struct subscribe_events_np *sen;
 	struct port_properties_np *ppn;
+	struct port_stats_np *psn;
 	struct mgmt_clock_description *cd;
 	switch (m->id) {
 	case TLV_CLOCK_DESCRIPTION:
@@ -391,6 +432,11 @@ static void mgt_pre_send(struct management_tlv *m, struct tlv_extra *extra)
 		ppn = (struct port_properties_np *)m->data;
 		ppn->portIdentity.portNumber = htons(ppn->portIdentity.portNumber);
 		break;
+	case TLV_PORT_STATS_NP:
+		psn = (struct port_stats_np *)m->data;
+		psn->portIdentity.portNumber =
+			htons(psn->portIdentity.portNumber);
+		break;
 	}
 }
 
@@ -517,6 +563,10 @@ static int org_post_recv(struct organization_tlv *org)
 			scaled_ns_n2h(&f->lastGmPhaseChange);
 			f->scaledLastGmPhaseChange = ntohl(f->scaledLastGmPhaseChange);
 			break;
+
+		case 2:
+			if (org->length + sizeof(struct TLV) != sizeof(struct msg_interval_req_tlv))
+				goto bad_length;
 		}
 	}
 	return 0;
@@ -544,6 +594,191 @@ static void org_pre_send(struct organization_tlv *org)
 	}
 }
 
+static int slave_delay_timing_data_post_revc(struct tlv_extra *extra)
+{
+	struct slave_delay_timing_data_tlv *slave_delay =
+		(struct slave_delay_timing_data_tlv *) extra->tlv;
+	size_t base_size = sizeof(slave_delay->sourcePortIdentity), n_items;
+	struct slave_delay_timing_record *record;
+
+	if (tlv_array_invalid(extra->tlv, base_size, sizeof(*record))) {
+		return -EBADMSG;
+	}
+	n_items = tlv_array_count(extra->tlv, base_size, sizeof(*record));
+	record = slave_delay->record;
+
+	NTOHS(slave_delay->sourcePortIdentity.portNumber);
+
+	while (n_items) {
+		NTOHS(record->sequenceId);
+		timestamp_net2host(&record->delayOriginTimestamp);
+		net2host64_unaligned(&record->totalCorrectionField);
+		timestamp_net2host(&record->delayResponseTimestamp);
+		n_items--;
+		record++;
+	}
+
+	return 0;
+}
+
+static void slave_delay_timing_data_pre_send(struct tlv_extra *extra)
+{
+	struct slave_delay_timing_data_tlv *slave_delay =
+		(struct slave_delay_timing_data_tlv *) extra->tlv;
+	size_t base_size = sizeof(slave_delay->sourcePortIdentity), n_items;
+	struct slave_delay_timing_record *record;
+
+	n_items = tlv_array_count(extra->tlv, base_size, sizeof(*record));
+	record = slave_delay->record;
+
+	HTONS(slave_delay->sourcePortIdentity.portNumber);
+
+	while (n_items) {
+		HTONS(record->sequenceId);
+		timestamp_host2net(&record->delayOriginTimestamp);
+		host2net64_unaligned(&record->totalCorrectionField);
+		timestamp_host2net(&record->delayResponseTimestamp);
+		n_items--;
+		record++;
+	}
+}
+
+static int slave_rx_sync_timing_data_post_revc(struct tlv_extra *extra)
+{
+	struct slave_rx_sync_timing_data_tlv *slave_data =
+		(struct slave_rx_sync_timing_data_tlv *) extra->tlv;
+	size_t base_size = sizeof(slave_data->sourcePortIdentity), n_items;
+	struct slave_rx_sync_timing_record *record;
+
+	if (tlv_array_invalid(extra->tlv, base_size, sizeof(*record))) {
+		return -EBADMSG;
+	}
+	n_items = tlv_array_count(extra->tlv, base_size, sizeof(*record));
+	record = slave_data->record;
+
+	NTOHS(slave_data->sourcePortIdentity.portNumber);
+
+	while (n_items) {
+		NTOHS(record->sequenceId);
+		timestamp_net2host(&record->syncOriginTimestamp);
+		net2host64_unaligned(&record->totalCorrectionField);
+		NTOHL(record->scaledCumulativeRateOffset);
+		timestamp_net2host(&record->syncEventIngressTimestamp);
+		n_items--;
+		record++;
+	}
+
+	return 0;
+}
+
+static void slave_rx_sync_timing_data_pre_send(struct tlv_extra *extra)
+{
+	struct slave_rx_sync_timing_data_tlv *slave_data =
+		(struct slave_rx_sync_timing_data_tlv *) extra->tlv;
+	size_t base_size = sizeof(slave_data->sourcePortIdentity), n_items;
+	struct slave_rx_sync_timing_record *record;
+
+	n_items = tlv_array_count(extra->tlv, base_size, sizeof(*record));
+	record = slave_data->record;
+
+	HTONS(slave_data->sourcePortIdentity.portNumber);
+
+	while (n_items) {
+		HTONS(record->sequenceId);
+		timestamp_host2net(&record->syncOriginTimestamp);
+		host2net64_unaligned(&record->totalCorrectionField);
+		HTONL(record->scaledCumulativeRateOffset);
+		timestamp_host2net(&record->syncEventIngressTimestamp);
+		n_items--;
+		record++;
+	}
+}
+
+static int unicast_message_type_valid(uint8_t message_type)
+{
+	message_type >>= 4;
+	switch (message_type) {
+	case ANNOUNCE:
+	case SYNC:
+	case DELAY_RESP:
+	case PDELAY_RESP:
+		return 1;
+	default:
+		return 0;
+	}
+}
+
+static int unicast_negotiation_post_recv(struct tlv_extra *extra)
+{
+	struct request_unicast_xmit_tlv *request;
+	struct ack_cancel_unicast_xmit_tlv *ack;
+	struct cancel_unicast_xmit_tlv *cancel;
+	struct grant_unicast_xmit_tlv *grant;
+	struct TLV *tlv = extra->tlv;
+
+	switch (tlv->type) {
+	case TLV_REQUEST_UNICAST_TRANSMISSION:
+		if (TLV_LENGTH_INVALID(tlv, request_unicast_xmit_tlv)) {
+			return -EBADMSG;
+		}
+		request = (struct request_unicast_xmit_tlv *) tlv;
+		if (!unicast_message_type_valid(request->message_type)) {
+			return -EBADMSG;
+		}
+		NTOHL(request->durationField);
+		break;
+	case TLV_GRANT_UNICAST_TRANSMISSION:
+		if (TLV_LENGTH_INVALID(tlv, grant_unicast_xmit_tlv)) {
+			return -EBADMSG;
+		}
+		grant = (struct grant_unicast_xmit_tlv *) tlv;
+		if (!unicast_message_type_valid(grant->message_type)) {
+			return -EBADMSG;
+		}
+		NTOHL(grant->durationField);
+		break;
+	case TLV_CANCEL_UNICAST_TRANSMISSION:
+		if (TLV_LENGTH_INVALID(tlv, cancel_unicast_xmit_tlv)) {
+			return -EBADMSG;
+		}
+		cancel = (struct cancel_unicast_xmit_tlv *) tlv;
+		if (!unicast_message_type_valid(cancel->message_type_flags)) {
+			return -EBADMSG;
+		}
+		break;
+	case TLV_ACKNOWLEDGE_CANCEL_UNICAST_TRANSMISSION:
+		if (TLV_LENGTH_INVALID(tlv, ack_cancel_unicast_xmit_tlv)) {
+			return -EBADMSG;
+		}
+		ack = (struct ack_cancel_unicast_xmit_tlv *) tlv;
+		if (!unicast_message_type_valid(ack->message_type_flags)) {
+			return -EBADMSG;
+		}
+		break;
+	}
+	return 0;
+}
+
+static void unicast_negotiation_pre_send(struct TLV *tlv)
+{
+	struct request_unicast_xmit_tlv *request;
+	struct grant_unicast_xmit_tlv *grant;
+
+	switch (tlv->type) {
+	case TLV_REQUEST_UNICAST_TRANSMISSION:
+		request = (struct request_unicast_xmit_tlv *) tlv;
+		HTONL(request->durationField);
+		break;
+	case TLV_GRANT_UNICAST_TRANSMISSION:
+		grant = (struct grant_unicast_xmit_tlv *) tlv;
+		HTONL(grant->durationField);
+		break;
+	case TLV_CANCEL_UNICAST_TRANSMISSION:
+	case TLV_ACKNOWLEDGE_CANCEL_UNICAST_TRANSMISSION:
+		break;
+	}
+}
+
 struct tlv_extra *tlv_extra_alloc(void)
 {
 	struct tlv_extra *extra = TAILQ_FIRST(&tlv_pool);
@@ -574,11 +809,10 @@ void tlv_extra_recycle(struct tlv_extra *extra)
 
 int tlv_post_recv(struct tlv_extra *extra)
 {
-	int result = 0;
-	struct management_tlv *mgt;
 	struct management_error_status *mes;
 	struct TLV *tlv = extra->tlv;
-	struct path_trace_tlv *ptt;
+	struct management_tlv *mgt;
+	int result = 0;
 
 	switch (tlv->type) {
 	case TLV_MANAGEMENT:
@@ -605,15 +839,15 @@ int tlv_post_recv(struct tlv_extra *extra)
 	case TLV_GRANT_UNICAST_TRANSMISSION:
 	case TLV_CANCEL_UNICAST_TRANSMISSION:
 	case TLV_ACKNOWLEDGE_CANCEL_UNICAST_TRANSMISSION:
+		result = unicast_negotiation_post_recv(extra);
 		break;
 	case TLV_PATH_TRACE:
-		ptt = (struct path_trace_tlv *) tlv;
-		if (path_length(ptt) > PATH_TRACE_MAX) {
-			ptt->length = PATH_TRACE_MAX * sizeof(struct ClockIdentity);
+		if (tlv_array_invalid(tlv, 0, sizeof(struct ClockIdentity))) {
+			goto bad_length;
 		}
 		break;
 	case TLV_ALTERNATE_TIME_OFFSET_INDICATOR:
-	case TLV_AUTHENTICATION:
+	case TLV_AUTHENTICATION_2008:
 	case TLV_AUTHENTICATION_CHALLENGE:
 	case TLV_SECURITY_ASSOCIATION_UPDATE:
 	case TLV_CUM_FREQ_SCALE_FACTOR_OFFSET:
@@ -622,6 +856,25 @@ int tlv_post_recv(struct tlv_extra *extra)
 	case TLV_PTPMON_RESP:
 		result = nsm_resp_post_recv(extra);
 		break;
+	case TLV_ORGANIZATION_EXTENSION_PROPAGATE:
+	case TLV_ENHANCED_ACCURACY_METRICS:
+	case TLV_ORGANIZATION_EXTENSION_DO_NOT_PROPAGATE:
+	case TLV_L1_SYNC:
+	case TLV_PORT_COMMUNICATION_AVAILABILITY:
+	case TLV_PROTOCOL_ADDRESS:
+		break;
+	case TLV_SLAVE_RX_SYNC_TIMING_DATA:
+		result = slave_rx_sync_timing_data_post_revc(extra);
+		break;
+	case TLV_SLAVE_RX_SYNC_COMPUTED_DATA:
+	case TLV_SLAVE_TX_EVENT_TIMESTAMPS:
+		break;
+	case TLV_SLAVE_DELAY_TIMING_DATA_NP:
+		result = slave_delay_timing_data_post_revc(extra);
+		break;
+	case TLV_CUMULATIVE_RATE_RATIO:
+	case TLV_PAD:
+	case TLV_AUTHENTICATION:
 	default:
 		break;
 	}
@@ -654,9 +907,11 @@ void tlv_pre_send(struct TLV *tlv, struct tlv_extra *extra)
 	case TLV_GRANT_UNICAST_TRANSMISSION:
 	case TLV_CANCEL_UNICAST_TRANSMISSION:
 	case TLV_ACKNOWLEDGE_CANCEL_UNICAST_TRANSMISSION:
+		unicast_negotiation_pre_send(tlv);
+		break;
 	case TLV_PATH_TRACE:
 	case TLV_ALTERNATE_TIME_OFFSET_INDICATOR:
-	case TLV_AUTHENTICATION:
+	case TLV_AUTHENTICATION_2008:
 	case TLV_AUTHENTICATION_CHALLENGE:
 	case TLV_SECURITY_ASSOCIATION_UPDATE:
 	case TLV_CUM_FREQ_SCALE_FACTOR_OFFSET:
@@ -665,6 +920,25 @@ void tlv_pre_send(struct TLV *tlv, struct tlv_extra *extra)
 	case TLV_PTPMON_RESP:
 		nsm_resp_pre_send(extra);
 		break;
+	case TLV_ORGANIZATION_EXTENSION_PROPAGATE:
+	case TLV_ENHANCED_ACCURACY_METRICS:
+	case TLV_ORGANIZATION_EXTENSION_DO_NOT_PROPAGATE:
+	case TLV_L1_SYNC:
+	case TLV_PORT_COMMUNICATION_AVAILABILITY:
+	case TLV_PROTOCOL_ADDRESS:
+		break;
+	case TLV_SLAVE_RX_SYNC_TIMING_DATA:
+		slave_rx_sync_timing_data_pre_send(extra);
+		break;
+	case TLV_SLAVE_RX_SYNC_COMPUTED_DATA:
+	case TLV_SLAVE_TX_EVENT_TIMESTAMPS:
+		break;
+	case TLV_SLAVE_DELAY_TIMING_DATA_NP:
+		slave_delay_timing_data_pre_send(extra);
+		break;
+	case TLV_CUMULATIVE_RATE_RATIO:
+	case TLV_PAD:
+	case TLV_AUTHENTICATION:
 	default:
 		break;
 	}
