@@ -42,30 +42,41 @@ int transport_recv(struct transport *t, int fd, struct ptp_message *msg)
 	return t->recv(t, fd, msg, sizeof(msg->data), &msg->address, &msg->hwts);
 }
 
-int transport_send(struct transport *t, struct fdarray *fda, int event,
-		   struct ptp_message *msg)
+int transport_send(struct transport *t, struct fdarray *fda,
+		   enum transport_event event, struct ptp_message *msg)
 {
 	int len = ntohs(msg->header.messageLength);
 
 	return t->send(t, fda, event, 0, msg, len, NULL, &msg->hwts);
 }
 
-int transport_peer(struct transport *t, struct fdarray *fda, int event,
-		   struct ptp_message *msg)
+int transport_peer(struct transport *t, struct fdarray *fda,
+		   enum transport_event event, struct ptp_message *msg)
 {
 	int len = ntohs(msg->header.messageLength);
 
 	return t->send(t, fda, event, 1, msg, len, NULL, &msg->hwts);
 }
 
-int transport_sendto(struct transport *t, struct fdarray *fda, int event,
-		     struct ptp_message *msg)
+int transport_sendto(struct transport *t, struct fdarray *fda,
+		     enum transport_event event, struct ptp_message *msg)
 {
 	int len = ntohs(msg->header.messageLength);
 
 	return t->send(t, fda, event, 0, msg, len, &msg->address, &msg->hwts);
 }
 
+int transport_txts(struct fdarray *fda,
+		   struct ptp_message *msg)
+{
+	int cnt, len = ntohs(msg->header.messageLength);
+	struct hw_timestamp *hwts = &msg->hwts;
+	unsigned char pkt[1600];
+
+	cnt = sk_receive(fda->fd[FD_EVENT], pkt, len, NULL, hwts, MSG_ERRQUEUE);
+	return cnt > 0 ? 0 : cnt;
+}
+
 int transport_physical_addr(struct transport *t, uint8_t *addr)
 {
 	if (t->physical_addr) {
